dnl ------------------------------ biosapi.S --------------------------------
dnl the original TUNES module to run on raw ISA PCs
dnl
dnl see ~/info/BIOS-API.nfo
dnl
dnl
Require({i386})


;;;;;;;;;;;;;;;;;;;;;;;;; Version identification ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({API_MAGIC},{MAGIC_PC_API})
define({API_NAME},{"OSLm i386 BIOS API"})


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({APISEG},	0x0000)		; Where the API is originally loaded
define({APIOFS},	PAGESIZE)	; The corresponding offset in PM.
define({IRQ_INT_BASE},	0x90)		; IRQ's are mapped at INT 90h
define({IDT_SIZE},	0x100)		; Is that enough ?
define({PM_STACK_SIZE},	0x1000)		; 4096 entries
define({GDT_SIZE},	0x200)		; Is that enough ?
define({COMMAND_LINE},	0x800)		; Where to put the command line
define({V86_SERVER_INT},0xFD)		; V86 server interrupt

ASCII_defs
Std_Struc
API_Struc

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Some options ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({__IO_Delay__},
	{locals({D1},{D2},{jmp D1}__nl__{D1:	jmp D2}__nl__{D2:})})dnl


;;;;;;;;;;;;;;;;;;;;;;;;; Standard PM Selectors ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Name			sel.	base lim.	description

;; These are valid even when paging is disabled
;; thus, linear memory must be identity-mapped to physical memory,
;; or at least the PMPART of the API
NULLSEG		=	0x0000	;0	0	invalid null segment
TEMPSEG		=	0x0008	;?	?	temporary
ALLMEMCS	=	0x0010	;0	4G	whole memory (code)
ALLMEMDS	=	0x0018	;0	4G	whole memory (data)
RMPARTCS	=	0x0020	;?	?	RM Part of the API (code), noPG
RMPARTDS	=	0x0028	;?	?	RM Part of the API (data), noPG
PMPARTCS	=	0x0030	;?	4G	PM Part of the API (code), noPG
PMPARTDS	=	0x0038	;?	4G	PM Part of the API (data), noPG

;; These are valid only when paging is enabled
;; thus, for linear address offset to be 4MB aligned,
;; the API must be double mapped up in this offset linear memory
KTSS		=	0x0040	;?	?	Kernel TSS, PG
V86TSS		=	KTSS	;?	?	V86 TSS, PG
;DummyTSS	=	0x0048	;?	?	dummy for no-save switches.

KL0CS		=	0x0080	;1G	4G	Kernel PL0 Code Segment
KL0DS		=	0x0088	;1G	4G	Kernel PL0 Data Segment
KL1CS		=	0x0091	;1G	4G	Kernel PL1 Code Segment
KL1DS		=	0x0099	;1G	4G	Kernel PL1 Data Segment
KL2CS		=	0x00A2	;1G	4G	Kernel PL2 Code Segment
KL2DS		=	0x00AA	;1G	4G	Kernel PL2 Data Segment
KL3CS		=	0x00B3	;1G	4G	Kernel PL3 Code Segment
KL3DS		=	0x00BB	;1G	4G	Kernel PL3 Data Segment
KL0CG		=	0x00C0	;?	?	Kernel PL0 Call Gate
KL1CG		=	0x00C9	;?	?	Kernel PL1 Call Gate
KL2CG		=	0x00D2	;?	?	Kernel PL2 Call Gate
KL3CG		=	0x00DB	;?	?	Kernel PL3 Call Gate

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; segmenting ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; .text is rmpart
; .data is pmpart
; .bss is pmpart udata
; after all, everything is .text, so as86 is not confused

declarediversion({__rminitcode__})dnl
declarediversion({__rminitdata__})dnl
declarediversion({__rmcode__})dnl
declarediversion({__rmdata__})dnl
declarediversion({__rmstringseg__})dnl
declarediversion({__rmudata__})dnl
declarediversion({__pminitcode__})dnl
declarediversion({__pminitdata__})dnl
declarediversion({__pmcode__})dnl
declarediversion({__pmdata__})dnl
declarediversion({__pmstringseg__})dnl
declarediversion({__symbolseg__})dnl
declarediversion({__pmudata__})dnl


define({pmaddr},	{pmofs+$1})
define({rmaddr},	{rmofs+$1})
define({initaddr},	{$1})


;;;;;;;;;;;;;;;;;;;;;;;;;; Initialize Segments ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define({__begin__},{
__pmcode__({
;;; SEGMENT: PM CODE
use32
pmcode_beg:
})
__pmdata__({
;;; SEGMENT: PM DATA
pmdata_beg:
dnl .text
})
__pmudata__({
pmudata_beg:
;;; SEGMENT: PM UDATA
newheap({u},pmudata_beg,0)
})
__pminitcode__({
;;; SEGMENT: PM INIT CODE
use32
pminitcode_beg:
})
__pminitdata__({
;;; SEGMENT: PM INIT DATA
pminitdata_beg:
})
__rmcode__({
;;; SEGMENT: RM CODE
dnl .text
use16
rmcode_beg:
})
__rmdata__({
;;; SEGMENT: RM DATA
rmdata_beg:
})
__rmudata__({
;;; SEGMENT: RM UDATA
dnl .text
rmudata_beg:
newheap({ru},rmudata_beg,0)
})
__rminitcode__({
;;; SEGMENT: RM INIT CODE
use16
rminitcode_beg:
})
__rminitdata__({
;;; SEGMENT: RM INIT DATA
dnl .text
rminitdata_beg:
})
__rmstringseg__({
;;; SEGMENT: RM STRING
rmstringseg_beg:
})
__pmstringseg__({
;;; SEGMENT: PM STRING
dnl .text
pmstringseg_beg:
})
__symbolseg__({
;;; SEGMENT: SYMBOL
symbolseg_beg:


})
})

define({rminit},{
	define({__codeseg__},{__rminitcode__($}{@)})
	define({__dataseg__},{__rminitdata__($}{@)})
	define({__udataseg__},{__rminitdata__($}{@)})
	define({__stringseg__},{__rminitdata__($}{@)})
	define({stringaddr},{rmaddr($}{1)})
})
define({pminit},{
	define({__codeseg__},{__pminitcode__($}{@)})
	define({__dataseg__},{__pminitdata__($}{@)})
	define({__udataseg__},{__pminitdata__($}{@)})
	define({__stringseg__},{__pminitdata__($}{@)})
	define({stringaddr},{pmaddr($}{1)})
})
define({pmpart},{
	define({__codeseg__},{__pmcode__($}{@)})
	define({__dataseg__},{__pmdata__($}{@)})
	define({__udataseg__},{__pmudata__($}{@)})
	define({__stringseg__},{__pmstringseg__($}{@)})
	define({stringaddr},{pmaddr($}{1)})
})
define({rmpart},{
	define({__codeseg__},{__rmcode__($}{@)})
	define({__dataseg__},{__rmdata__($}{@)})
	define({__udataseg__},{__rmudata__($}{@)})
	define({__stringseg__},{__rmstringseg__($}{@)})
	define({stringaddr},{rmaddr($}{1)})
})


define({__end__},{
__pmcode__(
{
.align 16
pmcode_end:
})
__pmdata__(
{
.align 16
pmdata_end:
})
__pmudata__(
{
pmudata_end = alallot(u,16,0)
})
__pminitcode__(
{
.align 16
pminitcode_end:
})
__pminitdata__(
{
.align 16
pminitdata_end:
})
__rmcode__(
{
.align 16
rmcode_end:
})
__rmdata__(
{
.align 16
rmdata_end:
})
__rmudata__(
{
rmudata_end = alallot(ru,16,0)
})
__rminitcode__(
{
.align 16
rminitcode_end:
})
__rminitdata__(
{
.align 16
rminitdata_end:
})
__rmstringseg__(
{
.align 16
rmstringseg_end:
})
__pmstringseg__(
{
.align 16
pmstringseg_end:
})
__symbolseg__(
{
.align 16
symbolseg_end:
})
;undiverting
undivert

rminitcode_len	= rminitcode_end - rminitcode_beg
rminitdata_len	= rminitdata_end - rminitdata_beg
rmcode_len	= rmcode_end - rmcode_beg
rmdata_len	= rmdata_end - rmdata_beg
rmudata_len	= rmudata_end - rmudata_beg
rmstringseg_len	= rmstringseg_end - rmstringseg_beg

pminitcode_len	= pminitcode_end - pminitcode_beg
pminitdata_len	= pminitdata_end - pminitdata_beg
pmcode_len	= pmcode_end - pmcode_beg
pmdata_len	= pmdata_end - pmdata_beg
pmudata_len	= pmudata_end - pmudata_beg
pmstringseg_len	= pmstringseg_end - pmstringseg_beg
symbolseg_len	= symbolseg_end - symbolseg_beg

pmpart_beg	= pmcode_beg
pmpart_end	= pmudata_end
pmpart_len	= pmpart_end - pmpart_beg
pmpart_actbeg	= pminitcode_beg
pmpart_actend	= symbolseg_end
pmpart_actlen	= pmpart_actend - pmpart_actbeg
pmpart_bsslen	= pmudata_len

rmpart_beg	= rmcode_beg
rmpart_end	= rmudata_end
rmpart_len	= rmpart_end - rmpart_beg
rmpart_actbeg	= rminitcode_beg
rmpart_actend	= rmstringseg_end
rmpart_actlen	= rmpart_actend - rmpart_actbeg

total_length	= rmpart_actlen + pmpart_actlen

pmofs	= -pmpart_end
rmofs	= 0x10000-rmpart_end


})

dnl ---------------- Some constants -----------------
dnl
APISIZE = total_length


define({rz__z},
    {zzz({
     ifelse({$1},,,
	{pushax
	 PUTS($1)
	 popax})
     call RM_DumpERegs
     ifelse({$2},,,
	{pushax
	 CGETC
	 popax})
     })})dnl




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Actual code ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({DEBUG})
define({__DEBUG__})


.org APIOFS

__OUT__({beginning}__nl__)
__begin__

Require({header})
Require({head})
Require({rmpart})
Require({pmpart})

zzz({
Require({debug})
Require({rmdebug})
})

Require({vp})



__OUT__({end reached}__nl__)
__end__
