dnl ----------------------=|=--- head.m4 ---=|=-----------------------------
dnl --=-- Initialization of the TUNES i386 BIOS API --=--
dnl
dnl
dnl Lots of info taken from the sources of the Linux console driver.
dnl
Provide({head},{
dnl
dnl


rminit

__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; API RM Entry Point ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

entry API_Entry_Point


API_Entry_Point:

; Called by the boot loader.
; must be far away from the top of conventional memory.
; * Entry:
;  CS,DS,ES,SS point to the API group of segments
;  the COMMAND_LINE is below APIOFS, the API offset.
;  SP is below the COMMAND_LINE. (!)Beware, it's less than 768 bytes long !
;  arguments are pushed on the stack: see below

; * Never exits, but launches TUNES.

pushdef({stringaddr},{$1})

	__("Initializing the BIOS-based API.",CRLF)


;;; Init Console

   ;; Define Video Mode: text mode 80x25, 16 colors, clear screen @ 0xB8000
	mov ax,#0x0003
	int #0x10

   ;; Print Init Message
	PUTS("Welcome to TUNES.",CRLF)

   ;; Set the keyboard repeat rate to the max
	mov ax,#0x0305
	clr(bx)
	int #0x16


;;; Get parameters from the LOADER
	; Get CPU type
	pop word [CPUType]
	zdwoz({"CPU:       0x"},[CPUType],{CRLF})

	; Get available conventional memory
	pop eax			; optimize for below
	mov [AvConvMem],eax
	zdwoz({"AvConvMem: 0x"},[AvConvMem],{CRLF})

	; Get total conventional memory
	pop dword [ConvMem]
	zdwoz({"ConvMem:   0x"},[ConvMem],{CRLF})

	; Get available/total extended memory
	pop dword [XtdMem]
	zdwoz({"XtdMem:    0x"},[XtdMem],{CRLF})

	; Get command line
	; (nothing to do; loader put it at COMMAND_LINE, zero-ended)
	zzz({
	  pushad
	  PUTS("Command line is:",CRLF)
	  mov si,#COMMAND_LINE
	  PUTS
	  call PrCRLF
	  popad
	})


;;; Move the Real-Mode Part to where it belongs
   ;; Compute the target location for RM Part.
	;mov eax,[AvConvMem]	; optimized above
	shr eax,#4
	sub ax,#0x1000		; 64 KB in para
   ;; Initialize some variables with it
	mov [RM_Part_seg],ax
	mov [RM_Target_seg],ax	; target when back into RM...
   ;; Do move
	mov si,#rmpart_actbeg
	mov es,ax
	mov di,#rmaddr(rmpart_actbeg)
	mov cx,#(rmpart_actlen+3)/4
	cld
	rep
	movsd

forget({
   ;; little test of iret(d)
	pushfd
	push dword cs
	push dword #test_iretd
	mov bp,sp
	iretd
test_iretd:
	sub bp,sp
})

;;; Continue up there...
   ;; use stack and data in rmpart
	;mov ax,[RM_Part_seg]
        mov ss,ax
	mov esp,#rmaddr(RMPartStackTop)
	mov ds,ax
popdef({stringaddr})
	push #0
	pop es
   ;; jump in RM Part...
	push ax
	push #rmaddr(UpThere)
	PUTS("Jumping up... ")
	retf

UpThere:
	PUTS("Here I am up in RM part !",CRLF)

;;; We go in protected mode as soon as possible:
;;; everything is *so* much easier from there,
;;; and we write everything for it anyway,
;;; so why write a crippled redundant real-mode version ?

;;; Initialize RM Part variables
   ;; compute RM and PM values for cr0
	mov eax,cr0
	mov [rmaddr(RM_cr0)],eax
	and eax,#~(CR0_PG|CR0_CD|CR0_NW|CR0_AM|CR0_WP)
	or al,#CR0_PE
	mov [rmaddr(PM_cr0)],eax

;;; Initialize PM Part variables
   ;; save real-mode IDT
	seg es
	sidt [rm_idt_address]
	zWZonez({"IDT: "},es,rm_idt_address,3,{CRLF})
   ;; ain't that all ?
	;.....

;;; Disable all interrupts
	cli
	__({"Disabling IRQs"},CRLF)
	call RM_Mask_IRQs
	__({"Disabling NMI"},CRLF)
	call RM_Disable_NMI
	__({"All interrupts disabled"},CRLF)

;;; Jump into the Protected Mode
	__("Let's jump into protected mode !",CRLF)
	br GoProtect




;;;;;;;;;;;; See below for protected mode code ;;;;;;;;;;;;;;;;;;;


;;;; Test code for R86/V86 reflectors...

V86_Test:
	STOP(es,0,40,22,'V,0xC)
	CGETC({{"Hello, V86 World !"},CRLF})
	STOP(es,0,41,22,'T,0xC)
	push cs
	call V86_EXIT

R86_Test:
	STOP(es,0,40,23,'R,0xA)
	pushing(es,{
	mov ax,#0x0003 ; clear screen
	int #0x10
	})
	CGETC({{"Hello, R86 World !"},CRLF})
	STOP(es,0,41,23,'T,0xA)
	push cs
	call R86_EXIT

RM_Return_Point:
; Nothing to restore: every was restored by the R86 handler !
; Show that we got there
	STOP(fs,0xB8000,0,9,'R,0xE)
	STOP(es,0,75,24,'R,0xC)		; 'R' for RealMode
	STOP(es,0,76,24,'e,0xC)		; 'e' for RealMode
; Try accessing ALLMEMDS through fs.
	STOP(fs,0xB8000,77,24,'A,0xC)	; 'A' for Allmem
	STOP(fs,0xB8000,78,24,'l,0xC)	; 'l' for Allmem
	STOP(fs,0xB8000,79,24,'l,0xC)	; 'l' for Allmem


; Print some stuff

	seg fs
	mov eax,[IRQ_INT_BASE*4]
	seg fs
	mov ebx,[IRQ_INT_BASE*4+8*4]

	rz__z
	STOP(fs,0xB8000,71,9,'a,0xB)	; another 'a'




; Restore Interrupt Controller
	STOP(es,0,58,9,'R,0x61)
	;call RM_Setup_RM_8259
	STOP(es,0,59,9,'8,0x61)
	;call RM_Flush_Interrupts
	STOP(es,0,70,9,'B,0xB)		; 'B' for Back

; Reenable Interruptions
	call RM_Unmask_IRQs
	call RM_Enable_NMI
	push cs
	pop ds
	STOP(es,0,71,9,'a,0xB)		; 'A' for Back


	__({"Back in real mode:"},CRLF)
	STOP(es,0,72,9,'c,0xB)		; 'c' for Back
	z_z

; That's all, folks
	STOP(es,0,73,9,'k,0xB)		; 'k' for Back
	CGETC({CRLF,"Hit any key to reboot."})
	STOP(es,0,74,8,'!,0xB)		; '!' for Back!
; Reboot !
	cli
	mov ax,#0x1234
	push #0
	pop es
	seg es
	mov [0x472],ax
	seg es
	mov [0x467],edx
	mov ax, #0x50f                ;shutdown
	out #0x70, al                 ;write to CMOS
__WaitOutReady1:	; Busy-wait until 8042 is ready for new command
	in al,#0x64	; read 8042 status byte
	test al,2	; Bit 1 of status indicates input buffer full
	jnz __WaitOutReady1
	mov al,#0xFE	; Pulse "reset" = 8042 pin 0
	out #0x64,al
	___die___	; It takes some time to reboot, so let's wait a bit.

  ;; Warm Reboot doesn't work on my T2400CT; Why ???
	seg fs
	mov word [0x472],#0x1234
	jmpi 0,-1

forget({
Test_Mode:
; This routine has been used instead of RM_Return_Point,
; to prove that returning in real mode from a 32 bit protected mode segment
; made the CPU land in such thing as a 32 bit real mode !!!
; To have it work, copy the GoReal procedure in a 32 bit segment,
; and have it jump into this routine.
; (beware: jump address should be word ofs,word 0,word seg)
; print 'M' indifferently in 32 bit and 16 bit modes
; In both modes, it will print a M white on black on the first character of
; the second line.
; In 32 bit mode, it will also print next to it a blinking {acute E}
; (which corresponds to the nop's 9090h value which is pushed then poped !!!)
	push #0xB800
	nop
	nop
	pop ds
	push #0x74D
	nop
	nop
	xor bx,bx
	push #160
	nop
	nop
	pop dx
	mov bl,dl
	mov di,bx
	pop word [bx]
	hlt		; halt, or the computer may reboot
			; before you see the result.
})
})



pminit

__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;; API PM Entry Point ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


PM_Entry_Point:
;;; Initialize Registers
   ;; clear AH and high of EAX
	clr(eax)
   ;; DS,ES: all memory
	mov al,#ALLMEMDS
	mov ds,ax
	mov es,ax
	mov fs,ax
	STOP(ds,0xB8000,78,00,'P,0x78)
	STOP(ds,0xB8000,79,00,'M,0x78)

   ;; compute the target location for PM Part.
	; (?) should we limit it to 16 MB, as linux allows ???
	; well, in that case, it'd be better to change the LOADER, I guess.
	; people would also add a handler for extra memory as buffer...
      ;; EBP is top of memory
	mov ebp,[XtdMem]
	add ebp,#0x100000 ; Xtdmem is above 1 MB...
      ;; for further paging, we must be 4KB aligned...
	and ebp,#PAGEMASK

   ;; Initialize SS:
   ;; same all memory segment, stack physically where it belongs.
	mov ss,ax
	lea esp,[ebp+pmaddr(__Stack_Top__)]

	;NB: I lost a lot of time finding a bug:
	;  ***addressing ebp uses ss***,
	;  but I had ss in the rmpart segment,
	;  so anytime I tried to use ebp-addressing,
	;  it just crashed (GPF) 8-( )-8

   ;; GS: original rmpart
	push #RMPARTDS
	pop gs

   ;; Once and for all, have direction flag default to up in PM
	cld


;;; Print "Welcome to "
	mov edi,#0xB8000+2*(00+0*80)
	mov eax,#0x07650757	; 'We'
	stosd
	mov eax,#0x0763076C	; 'lc'
	stosd
	mov eax,#0x076D076F	; 'om'
	stosd
	mov eax,#0x07200765	; 'e '
	stosd
	mov eax,#0x076F0774	; 'to'
	stosd


;;; Move the Protected-Mode Part to where it belongs: up extended memory.
   ;; do move
	mov esi,#pmpart_actbeg
	lea edi,[ebp+pmaddr(pmpart_actbeg)]
	mov ecx,#(pmpart_actlen+3)/4
	rep
	movsd
   ;; clear the zone containing uninitialized PM part data
	clr(eax)
	mov ecx,#(pmpart_bsslen+3)/4
	rep
	stosd

;;; Initialize the GDT
   ;; Copy the original GDT
	mov esi,#Orig_GDT
	lea edi,[ebp+pmaddr(__GDT__)]
	mov ecx,#Orig_GDT_Len/4
	rep
	movsd
   ;; Modify the PMPART entries
	mov ax,#0xD09A
	mov ebx,ebp
	mov ecx,#0xFFFFF
	lea edx,[ebp+pmaddr(__GDT__+PMPARTCS)]
	call SetDescriptor
	add edx,#PMPARTDS-PMPARTCS
	mov al,#0x92
	call SetDescriptor
   ;; Modify the RMPART entries
	mov ebx,[RM_Part_seg]
	shl ebx,4
	mov ah,#0x10
	shr ecx,4
	add edx,#RMPARTDS-PMPARTDS
	call SetDescriptor
	mov al,#0x9A
	add edx,#RMPARTCS-RMPARTDS
	call SetDescriptor
   ;; Set the GDT variable in RM part
	lea eax,[edx-RMPARTCS]
	mov [ebx+rmaddr(PM_GDT_base)],eax
	mov word [ebx+rmaddr(PM_GDT_lim)],#8*GDT_SIZE-1
   ;; Re-load the GDT
	lgdt [ebx+rmaddr(PM_GDT_address)]

;;; Jump to there
	STOP(ds,0xB8000,78,01,'U,0x78)
	jmpi pmaddr(UpTherePM),PMPARTCS
UpTherePM:
	STOP(ds,0xB8000,79,01,'P,0x78)
   ;; Have DS, ES, and SS point to the new PMPART
	push #PMPARTDS
	pop ds
	push ds
	pop es
	push ds
	pop ss
	sub esp,ebp
   ;; Initialize some variables
	mov eax,ebx
	mov [pmaddr(RM_Part_Phys)],eax
	add [pmaddr(V86_REQUEST_Adjust-4)],eax
	add [pmaddr(V86_EXIT_Adjust-4)],eax
	sub eax,ebp
	mov [pmaddr(RM_Part_Addr)],eax
	mov [eax+rmaddr(PM_Target_ofs)],#pmaddr(Return_From_R86)
	add [pmaddr(pm_idt_base)],ebp
	sub [pmaddr(Video_mem_base)],ebp
   ;; Now, have ebp point to the beginning of memory in PMPARTDS
	mov ebx,ebp
	neg ebx
	mov [pmaddr(PhysicalMem)],ebx

;;; Proudly print " Tunes !"
	lea edi,[ebx+0xB8000+2*(10+0*80)]
	mov eax,#0x0F540F20	; ' T'
	stosd
	mov eax,#0x0F6E0F75	; 'un'
	stosd
	mov eax,#0x0F730F65	; 'es'
	stosd
	mov eax,#0x07210720	; ' !'
	stosd
	mov eax,#0x07200720
	mov ecx,#(80*25-18)/2
	rep
	stosd


;;; Setup TSSs
	STOP(fs,0xB8000,0,24,'T,0x0C)
   ;; Entry for "Kernel" TSS
	mov edx,#pmaddr(__GDT__)+KTSS
	clr(eax)	; mov ax,#0x0089
	mov al,#0x89
	clr(ecx)	; mov ecx,#0xFFFF
	dec cx
	add ebx,#pmaddr(K_TSS)
	call SetDescriptor
	mov [pmaddr(K_TSS+TSS32_ESP_CPL0)],#pmaddr(__Safe_Stack_Top__)
	mov [pmaddr(K_TSS+TSS32_SS_CPL0)],#PMPARTDS
forget({
   ;; Entry for Dummy TSS
	add edx,#DummyTSS-KTSS
	add ebx,#Dummy_TSS-K_TSS
	call SetDescriptor
   ;; Entry for V86 TSS
	mov edx,#V86TSS-KTSS
	add ebx,#V86_TSS-K_TSS
	call SetDescriptor
})
   ;; that's all for now.
	STOP(fs,0xB8000,1,24,'S,0x04)


;;; Setup Interrupts
	STOP(fs,0xB8000,0,24,'I,0x0B)
   ;; Redirect Hardware Interrupts
	call Setup_PM_8259
	STOP(fs,0xB8000,1,24,'H,0x03)
   ;; Initialize RM IDT:
	call Init_RM_IDT
	STOP(fs,0xB8000,2,24,'R,0x03)
   ;; Initialize ReflINT V86INT:
	mov eax,[pmaddr(RM_Part_seg)]
	mov [pmaddr(ReflINT_V86INT+V86INT_SS)],eax
	mov [pmaddr(ReflINT_V86INT+V86INT_ESP)],#rmaddr(RMPartStackTop)
	STOP(fs,0xB8000,3,24,'V,0x03)
   ;; Initialize PM IDT:
      ;; Move the parts of original IDT: exception handlers
	mov esi,#pmaddr(Orig_IDT_exc)
	mov edi,#pmaddr(__IDT__)
	mov ecx,#Orig_IDT_exc_Len/4
	rep
	movsd
	STOP(fs,0xB8000,4,24,'e,0x03)
      ;; Move the parts of original IDT: interrupt handlers
	add edi,#IRQ_INT_BASE*8-Orig_IDT_exc_Len
	mov ecx,#Orig_IDT_irq_Len/4
	rep
	movsd
	STOP(fs,0xB8000,5,24,'i,0x03)
   ;; Load newly initialize IDT:
	lidt [pmaddr(pm_idt_address)]
	STOP(fs,0xB8000,6,24,'l,0x03)

forget({
;;; Setup Real RM as supervised by PM
   ;; Initialize RM Interceptor IDT
	; if in "intercept all RM interrupts" mode,
	; all interrupts point to the same place
	; except that the low byte of cs indicates the interrupt number.
	mov ebx,[pmaddr(RM_Part_Addr)]
	mov eax,[pmaddr(RM_Part_seg)]
	mov [ebx+rmaddr(Actual_INT_Interceptor_seg)],ax
	add eax,eax
	lea eax,[eax*8+rmaddr(INT_Interceptor)-0x2000] ;8*2=16 !!!
	mov edx,eax
	and edx,#0x00FFF
	and eax,#0xFF000
	shl eax,#12
	lea eax,[eax+edx+0x2000]
	mov edi,#pmaddr(__RM_IRQ__)
	mov cx,#256
    Init_RM_Interceptor_IDT_0:
	stosd
	add eax,#0x0000FFF0
	loop Init_RM_Interceptor_IDT_0
	STOP(fs,0xB8000,6,24,'R,0x03)
})




;;; Initialize a stream input to what follows the API in the image file...
	;......



;;; Reenable interrupts
	call Enable_NMI
	call Unmask_IRQs
	call Flush_Interrupts
	sti
	STOP(fs,0xB8000,70,24,'I,0x0C)


;;; Here we are (Pphew !)
;;; Jump into LLL code
	STOP(fs,0xB8000,71,24,'J,0x0C)
	;......
	STOP(fs,0xB8000,72,24,'K,0x0C)


;;; Do some silly routine testing
	mov [pmaddr(DumpLine)],#3
;forget({
	mov edi,[pmaddr(Video_mem_base)]
	add edi,#80*2*19
	mov word [edi],# ('#)+0x8F00
	STOP(fs,0xB8000,1,19,'@,0x8F)
	lsa(esi,"Testing DumpString",0)
	mov ah,#0x07
	call DumpString
	mov eax,#1
	mov ebx,#2
	mov ecx,#3
	mov edx,#4
	mov esi,#5
	mov edi,#6
	mov ebp,#7
	z__z
	STOP(fs,0xB8000,18,19,'!,0x8F)
;})
	sub esp,#6
	sgdt [esp]
	clr(ebp)
	pop bp
	pop edi
	lea edx,[edi+RMPARTCS]
	add edx,[pmaddr(PhysicalMem)]
	call GetDescriptor
	mov esi,[pmaddr(RM_Part_seg)]
	z__z
zzz({
})

	mov ebp,#0xDEADBEEF

	STOP(fs,0xB8000,73,24,'T,0x0E)
	mov ebx,#pmaddr(Test_V86INT)
	mov eax,[pmaddr(RM_Part_seg)]
	mov [ebx+V86INT_CS],eax
	mov [ebx+V86INT_DS],eax
	mov [ebx+V86INT_ES],#0xB800
	mov [ebx+V86INT_SS],eax
	STOP(fs,0xB8000,71,24,'8,0x0E)
	STOP(fs,0xB8000,72,24,'6,0x0E)


	STOP(fs,0xB8000,39,0,'I,0x0F)
	int #IRQ_INT_BASE
	STOP(fs,0xB8000,40,0,'0,0x0F)

	cli
 
	mov ecx,10
	call sleep


forget({
;;; Test Call_INT_V86
	STOP(fs,0xB8000,64,23,'I,0x0E)
	mov dword [ebx+V86INT_EAX],#0x3
	mov cl,#0x10
	call DumpV86INT
	call Call_INT_V86
	STOP(fs,0xB8000,65,23,'X,0x0E)
	mov [pmaddr(DumpLine)],#3
	mov eax,[ebx+V86INT_EAX]
	mov ecx,#pmaddr(Test_V86INT)
	mov edi,#0x86118611
	z__z
	STOP(fs,0xB8000,66,23,'#,0x0E)

	___die___

;;; Test Call_V86
	STOP(fs,0xB8000,72,23,'V,0x0E)
	call DumpV86INT
	mov dword [ebx+V86INT_EIP],#rmaddr(V86_Test)
	mov dword [ebx+V86INT_EFLAGS],#0x00027000
	call Call_V86
	STOP(fs,0xB8000,73,23,'X,0x0E)
	mov eax, [ebx+V86INT_EAX]
	mov edi,#0x86008600
	mov [pmaddr(DumpLine)],#3
	z__z
	STOP(fs,0xB8000,74,23,'#,0x0E)
})


;;; Test Call_INT_R86
	STOP(fs,0xB8000,60,23,'I,0x0E)
	mov dword [ebx+V86INT_EAX],#0x3
	mov cl,#0x10
	call DumpV86INT
	call Call_INT_R86
	STOP(fs,0xB8000,61,23,'X,0x0E)
	mov [pmaddr(DumpLine)],#3
	call DumpV86INT
	mov eax,[ebx+V86INT_EAX]
	mov ecx,#pmaddr(Test_V86INT)
	mov edi,#0x11861186
	z__z
	STOP(fs,0xB8000,62,23,'#,0x0E)


;;; Test Call_R86
	STOP(fs,0xB8000,68,23,'R,0x0E)
	mov dword [ebx+V86INT_EIP],#rmaddr(R86_Test)
	mov eax,[pmaddr(RM_Part_seg)]
	mov [ebx+V86INT_CS],eax
	call DumpV86INT
	call Call_R86
	STOP(fs,0xB8000,69,23,'X,0x0E)
	mov eax,[ebx+V86INT_EAX]
	mov ecx,#pmaddr(Test_V86INT)
	mov [pmaddr(DumpLine)],#3
	mov edi,#0x00860086
	z__z
	STOP(fs,0xB8000,70,23,'#,0x0E)


;;; Exit back to real mode
   ;; Bye bye !

	STOP(fs,0xB8000,78,24,'B,0x0C)

	cli
	call Mask_IRQs
	call Disable_NMI
	call Setup_RM_8259
	call Flush_Interrupts
	call DumpV86INT
	mov ecx,100
	call sleep
	call Call_R86

   ;; Error ?
	STOP(fs,0xB8000,79,24,'E,0x8F)
	___die___ ; This point should never be reached !

Test_V86INT:
	.long 1,2,3,4,5,6,7,8		; (pushad)
	.long 0,0			; EIP,CS
	.long 0x00007046		; EFLAGS
	.long rmaddr(RMPartStackTop),0	; ESP,SS
	.long 0xB800,0,0,0		; ES,DS,FS,GS

})

__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;; Initialization Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.align 4
Orig_GDT:
 DefDescriptor(0,0x00000,0x00,0x00)				; NULL
 DefDescriptor(0,0xFFFFF,0x9A,0x50)				; TEMPSEG
 DefDescriptor(0,0xFFFFF,0x9A,0xD0)				; ALLMEMCS
 DefDescriptor(0,0xFFFFF,0x92,0xD0)				; ALLMEMDS
 DefDescriptor(rmpart_end-0x10000,0xFFFF,0x9A,0x10)		; RMPARTCS
 DefDescriptor(rmpart_end-0x10000,0xFFFF,0x92,0x10)		; RMPARTDS
 DefDescriptor(pmpart_end,0xFFFFF,0x9A,0xD0)			; PMPARTCS
Orig_GDT_end:
Orig_GDT_Len = Orig_GDT_end-Orig_GDT

Orig_IDT_exc:
irp({exc_handler},
{Division_Error_Trap,
 Single_Step_Trap,
 NMI_Trap,
 BreakPoint_Trap,
 Overflow_Trap,
 Out_of_Bound_Trap,
 Invalid_Opcode_Trap,
 Emulate_Coproc_Trap,
 Double_Fault_Stop,
 Coproc_Segment_Stop,
 Invalid_TSS_Fault,
 Segment_Not_Present_Fault,
 Stack_Fault,
 General_Protection_Fault,
 Page_Fault,
 Reserved_0F_Fault,
 Coprocessor_Error_Fault,
 Alignment_Check_Fault},
{DefGate(PMPARTCS,pmaddr(exc_handler),	0,0x8E)
})
Orig_IDT_exc_end:
Orig_IDT_exc_Len = Orig_IDT_exc_end - Orig_IDT_exc

;Orig_IDT_serv:
;Orig_IDT_serv_end:
;Orig_IDT_serv_Len = Orig_IDT_serv_end - Orig_IDT_serv

Orig_IDT_irq:
  DefGate(PMPARTCS,pmaddr(IRQ0_Timer),		0,0x8E)	; 0
  DefGate(PMPARTCS,pmaddr(IRQ1_Keyboard),	0,0x8E)	; 1
  DefGate(PMPARTCS,pmaddr(IRQ2_Cascade),	0,0x8E)	; 2
  DefGate(PMPARTCS,pmaddr(IRQ3),		0,0x8E) ; 3
  DefGate(PMPARTCS,pmaddr(IRQ4),		0,0x8E) ; 4
  DefGate(PMPARTCS,pmaddr(IRQ5),		0,0x8E) ; 5
  DefGate(PMPARTCS,pmaddr(IRQ6),		0,0x8E) ; 6
  DefGate(PMPARTCS,pmaddr(IRQ7),		0,0x8E) ; 7
  DefGate(PMPARTCS,pmaddr(IRQ8_RTC),		0,0x8E) ; 8
  DefGate(PMPARTCS,pmaddr(IRQ9),		0,0x8E) ; 9
  DefGate(PMPARTCS,pmaddr(IRQ10),		0,0x8E) ; A
  DefGate(PMPARTCS,pmaddr(IRQ11),		0,0x8E) ; B
  DefGate(PMPARTCS,pmaddr(IRQ12),		0,0x8E) ; C
  DefGate(PMPARTCS,pmaddr(IRQ13),		0,0x8E) ; D
  DefGate(PMPARTCS,pmaddr(IRQ14),		0,0x8E) ; E
  DefGate(PMPARTCS,pmaddr(IRQ15),		0,0x8E) ; F
Orig_IDT_irq_end:
Orig_IDT_irq_Len = Orig_IDT_irq_end - Orig_IDT_irq
})

})dnl --- end of head.m4
dnl
