dnl --=|=-- i386.m4 --=|=--
dnl --=-- This is the m4 include file to expand generic programs --=--
dnl --=-- into as86 sources for the i386 subproject --=--
dnl
dnl See ~moose/info/i386impl.info for more info about this.
dnl
dnl
dnl
Provide({i386},{
dnl
Require({as86})
Require({version})
Require({magic})
dnl
dnl
dnl --- Structures ---
dnl
define({API_Struc},{
Struc({AS},{
; Signature_Struc
  _long_(Magic_Number)		; magic number from magic file
  _byte_(Object_Name,20)
  _byte_(Version,8)
})
dnl
Struc({AH},{
; API_Header_Struc
; size is 32. offset is 32 from beginning of API file
; (32 first bytes are signature and version)
  _long_(TotalLength)dnl	total length of the file
  _long_(EntryPoint)dnl		address of object to return
  _long_(Offset)dnl		offset where to load it
  _long_(HeaderSize)dnl		size of header fields in fil
  _long_(ExpandedSize)dnl	size when expanded to memory
  _long_(pad)dnl		fill in
  _long_(CheckSum)dnl		checksum
  _long_(LongCheckSum)dnl	long checksum
})dnl
dnl
dnl ----- Aerial (extra-light) threads -----
Struc({AT},{dnl
dnl AerialThread	; These threads are lighter than light: Aerial !
  _long_(SAVE_EBX)dnl	initial Top of Stack
  _long_(SAVE_ESI)dnl	Data Stack Pointer
  _long_(SAVE_EDI)dnl	Return Stack Pointer
  _long_(SAVE_EBP)dnl	Those two are reserved for the
  _long_(SAVE_ESP)dnl	threading mechanism.
  _long_(SAVE_RESP)dnl	ESP save when going ret-threaded
  _long_(STATE)dnl	flags used for thread sync'
  _long_(MAX_ESI)dnl	bottom of data stack
  _long_(MAX_EDI)dnl	bottom of return stack
  _long_(LINKBACK)dnl	Link Back to thread object
})dnl
dnl
})dnl
dnl
dnl
dnl --- Some values ---
dnl
define({BYTEBITS},8)dnl		bits in a byte (smallest addressable value)
define({CHARABITS},0)dnl	bits of char in address
define({WORDABITS},1)dnl	bits of word in address
define({CELLABITS},2)dnl	bits of cell in address
define({PAGEABITS},12)dnl	bits of page in address
dnl
define({CELL},	CELLSIZE)dnl		a cell is 4 bytes long
define({MAX_INT},eval(2**31-1))dnl	MAX integer
define({MIN_INT},eval(-2**31))dnl	MIN integer
dnl
dnl
dnl
dnl
dnl
dnl ----- abstraction of the i386 instruction set... -----
dnl
define({___CPU___},{i386})dnl
dnl
forget({
Registers:
* caller-saved: eax,ecx,edx
* callee-saved: ebx,esi,edi,ebp,esp,cs,ds,es,ss,fs,gs

*  eax,edx	scratch; not saved upon a thread switch.
*  ecx		scratch; [OPT: callee-saved, should be constant zero ?]
*  ebx		buffers top of stack; saved upon a thread switch.
*  edi		data stack pointer (growing down)
		If the stack is declared garbage-collection aware
		all values on the stack may be corrected according to the
		garbage collection on a "PAUSE", a "YIELD", a memory
		allocation call or any such thing.
		 And the thread will be unavailable during garbage
		collection. Else, no relocatable address may be remembered
		from a previous yield (), and none may be used at all if the
		thread is preemptible.
		Due to buffering, the stack must contain a ghost element
		to be loaded into ebx when empty.
*  esp		local return stack pointer (growing down),
		also heap pointer for non-real-time threads (?)
*  ebp		frame on the heap
*  esi		points on current instruction to execute...
*  cs,ds,es,ss	point to the PMPART selectors
*  fs		points to the ALLMEM selector
*  gs		points to the RMPART selector


* Values:
 - from -1GB to +1GB-1
 - else, should not survive a PAUSE.
* Pointers:
 - from +1GB to +3GB-1
 - else, not relocatable.
})
dnl
define({_cell_},	{_long_($@)})dnl
dnl
define({defcell},	{.long $1}__nl__)dnl	define a cell
define({defword},	{.word $1}__nl__)dnl	define a cell
define({defbyte},	{.byte $1}__nl__)dnl	define a cell
define({DefString},
	{DefCountedString({$1},quotestring({$2}))})dnl
dnl
dnl
dnl --- Define Pseudo-ASM syntax stuff ---
dnl
define({Terminator},{}__nl__{})dnl
define({Separator},{})dnl
define({Equate},{{$1} = {$2}})dnl
dnl
dnl
dnl --- Define Pseudo-ASM registers ---
dnl
define({PC},		{esi})dnl	program counter
define({PCINC},		{4})dnl	 	program counter increment
define({DSP},		{edi})dnl	data stack pointer
define({MDSINC},	{4})dnl		increment for data stack -- grow down
define({TOS},		{ebx})dnl	buffered top of stack
define({TOSL},		{bx})dnl	its lowest word
define({TOSLL},		{bl})dnl	its lowest byte
define({TOSLH},		{bh})dnl	its second byte
define({RSP},		{esp})dnl	return stack pointer
define({MRSINC},	{4})dnl		increment for return stack -- grow down
dnl
define({FRAME},		{ebp})dnl	a frame pointer for heap manipulation
define({R0},		{eax})dnl	a scratch register
define({R0L},		{ax})dnl	its lowest word
define({R0LL},		{al})dnl	its lowest byte
define({R0LH},		{ah})dnl	its second byte
define({R1},		{edx})dnl	second scratch register
define({COUNT},		{ecx})dnl	count register
define({COUNTLO},	{cl})dnl	count register -- low byte
define({INSTRTBL},	{InstrT})dnl	table for bytecode threading
dnl
dnl
dnl --- Define Pseudo-ASM pointers ---
dnl
define({point0},{[ifelse({$2},{},{{$1}},{$2},{0},{{$1}},{{$1+$2*CELL}})]})dnl
define({point},		{dword point0({$1},{$2})})dnl
define({pointb},	 {byte point0({$1},{$2})})dnl
define({pointw},	 {word point0({$1},{$2})})dnl
define({bytecast},	{byte $1})dnl
define({wordcast},	{word $1})dnl
define({cellcast},	{dword $1})dnl
dnl
dnl
dnl --- Define Pseudo-ASM Instructions ---
dnl
define({__Mov__},	{mov $1,$2})dnl
define({__Xchg__},	{xchg $1,$2})dnl
define({Clr},		{clr($1)})dnl
dnl
define({Add},		{add $1,$2})dnl
define({AddIm},		{ifelse({isreg($1)},{1},{lea $1,[$1+$2]},
				{Add($1,$2)})})dnl
define({Inc},		{inc $1})dnl
define({Dec},		{dec $1})dnl
define({Sub},		{sub $1,$2})dnl
define({Neg},		{neg $1})dnl
dnl
define({Not},		{not $1})dnl
define({And},		{and $1,$2})dnl
define({Or},		{or $1,$2})dnl
define({Xor},		{xor $1,$2})dnl
dnl
define({__Shr__},	{shr $1,$2})dnl
define({__Shl__},	{shl $1,$2})dnl
define({BSwapR0},	{bswaperx(a)})dnl
dnl
dnl
dnl --- Branching ---
dnl
define({Goto},		{br $1})dnl
define({Call},		{call $1})dnl
define({Return},	{ret})dnl
dnl
define({negcond},	{ncc($@)})dnl
define({Branch},
{ifelse(
    {$1},{cxz},{jcxz $4},
    {$1},{testz},{test $2,$3
		  jz $4},
    {$1},{testnz},{test $2,$3
		  jnz $4},
    {$3},{0},{setf($2)}__nl__{j$1 $4},
    {cmp $2,$3
     j$1 $4})
})dnl
dnl
define({Loopno0},
    {pushdef({beg_of_loop},newlabel)dnl
     pushdef({end_of_loop},newlabel)dnl
   beg_of_loop:
     $1
     loop beg_of_loop
   end_of_loop:
     popdef({beg_of_loop})dnl
     popdef({end_of_loop})dnl
})dnl
define({Loop},
    {pushdef({beg_of_loop},newlabel)dnl
     pushdef({end_of_loop},newlabel)dnl
     jecxz end_of_loop
   beg_of_loop:
     $1
     loop beg_of_loop
   end_of_loop:
     popdef({beg_of_loop})dnl
     popdef({end_of_loop})dnl
})dnl
dnl
dnl
dnl   ID: make the __IF__ etc macros in as86.m4 more generic
dnl and use that here.
dnl we will assume that all assembler have test instruction
dnl followed by branch/loop/whatever instruction, with possible
dnl special case for
dnl test to zero.
dnl the _comparenum_ macro compares two numbers
dnl the optional _compare0_ macro compares a number to zero
dnl the ....
dnl   I should really document the vp somewhere...
dnl
dnl
dnl --- absolute call/jump words
dnl
define({jmpa},{
ifelse({$2},,
	{push #$1
	 ret},
	{mov $2,#$1
	 jmp $2}
)})dnl
define({calla},{
ifelse({$2},,
	{mov eax,#$1
	 call eax},
	{mov $2,#$1
	 call $2}
)})dnl
dnl
dnl
dnl --- I/O ---
dnl
define({IsImmByte},
	{ifelse({IsNumber({$1})},{},{0},{eval(($1<256)&&($1>=0))})})dnl
defdef({__IO_Delay__},
	{locals({D1},{D2},{jmp D1}__nl__{D1:	jmp D2}__nl__{D2:})})dnl
define({__outx_dx__},
{Mov(dx,{$2})
Mov({$1},{$3})
out dx,$1
__IO_Delay__})dnl
define({__outx_xx__},
{Mov({$1},{$3})
out $2,$1
__IO_Delay__})dnl
define({outb},{
ifelse({IsImmByte({$1})},1,
		{__outx_xx__(al,$1,{$2})},
		{__outx_dx__(al,$1,{$2})})})dnl
define({outw},{
ifelse({IsImmByte({$1})},1,
		{__outx_xx__(ax,$1,{$2})},
		{__outx_dx__(ax,$1,{$2})})})dnl
define({outl},{
ifelse({IsImmByte({$1})},1,
		{__outx_xx__(eax,$1,{$2})},
		{__outx_dx__(eax,$1,{$2})})})dnl
dnl
dnl
dnl
dnl
dnl --- Register Data Conventions ---
dnl
define({PopReturn},	{ifelse({$1},,{AddIm(RSP,MRSINC)},{pop $1})})dnl
define({PushReturn},	{push $1})dnl
define({},		{})dnl
define({},		{})dnl
define({},		{})dnl
dnl
dnl
dnl --- Calling Convention ---
dnl
define({Call_LLL},
{irp(X,reverse(shift($@)),{PushData(X)})
locals({Sequel},{T1},{
PushReturn(PC)
call T1
T1:
add [RSP],Sequel-T1
lea PC,[RSP+RSINC]
Goto($1)
Sequel:
PopReturn(PC)
})})dnl
dnl
dnl
dnl
dnl ----- Defining FORTH words -----
dnl
dnl --- defining a counted string
dnl
define({DefCountedString},
{defstring({$1},{shift($@)},
{.byte strlen-1
})})dnl
dnl
dnl
dnl ----- defining a FORTH word -----
dnl
define({StringAddress},{pmaddr({$1})})dnl
define({SymbolFlags},{ifelse({$1},{},0,{{$1}})})dnl
define({SymbolValue},{pmaddr({$1})})dnl
dnl
dnl --- flags for FORTH words ---
define({FW_IMMED},	0x80)dnl	word is immediate
define({FW_INLINE},	0x01)dnl	word can be inlined
dnl
dnl
dnl
forget({
;------------------------- Signature --------------------------
;
; The first 32 bytes are reserved for the signature
; The first 20 bytes may be searched for to find the Kernel.
; The 12 remaining bytes may be used for version identification.
;
  .org 0
   MagicNumber		.long	0x01824B26FH
dnl	 big-endian 	386	i	L	S	O
        ;       00  0110000010	01001	01100	10011	01111
			dnl	 5 bit code for 'OSL' big-endian,
			dnl	 5 bit code for i, 10-bit code for 386
			dnl	 two clear bits.
   Signature		.ascii	"OSLm Kernel i386 v0.00.01"
			.byte	0

;------------------------ Sizes -------------------------------
  ;; (!) The first 32 bytes were reserved for the signature.
  .org 32
			dd	TotalKernelFileLength
			dd	KernelOffset
			dd	KInitJumpPoint
			dd	size KHeader
})dnl
dnl
dnl
dnl
dnl ----- Definitions -----
define({DefCompareWord0},
{DefWord({$1},{$2},{$3})
   Enter
   test TOS,TOS
   ifelse(
	{$5},{nz},
	{ neg TOS	; CY iff NZ
	  sbb TOS,TOS },
	{$5},{nz},
	{ neg TOS	; CY iff NZ
	  cmc
	  sbb TOS,TOS },
	{$5},{s},
	{ sar TOS,31 },
	{$5},{ns},
	{ sar TOS,31
	  not TOS },
	{
	  set$5 TOSLL
	  movsx TOS,TOSLL
	})
   Next
})dnl
dnl
dnl
define({DefCompareWord},
{DefWord({$1},{$2},{$3})
   Enter
   ifelse(
	{$5},{nz},
	{ sub TOS,NOS
	  neg TOS	; CY iff NZ
	  sbb TOS,TOS },
	{$5},{z},
	{ sub TOS,NOS
	  neg TOS	; CY iff NZ
	  cmc
	  sbb TOS,TOS },
	{$5},{b},
	{ cmp NOS,TOS
	  sbb TOS,TOS },
	{$5},{ae},
	{ cmp NOS,TOS
	  sbb TOS,TOS },
	{$5},{a},
	{ cmp TOS,NOS
	  sbb TOS,TOS },
	{$5},{be},
	{ cmp TOS,NOS
	  sbb TOS,TOS },
	{$5},{ge},
	{ sub TOS,NOS ; beware: reverse comparison
	  sar TOS,31 },
	{$5},{lt},
	{ sub TOS,NOS ; beware: reverse comparison
	  sar TOS,31
	  not TOS },
	{ cmp NOS,TOS
	  set$5 TOSLL
	  movsx TOS,TOSLL
	})
   Nip
   Next
})dnl
dnl
dnl
dnl --- Deduce defaults from these ---
Require({defaults})
dnl
dnl
})dnl ---- end of i386.m4 ----
dnl
