#!/bin/bash


############################## CONFIGURATION #################################

config () {
  LOADMODE=C
  BLOCKSIZE=512

  return 0		# The rest is now done from the Makefile
}



############################### IO ROUTINES ##################################

TMP=/tmp/tmp-moose.install.sh-$HOST-$$-


DD ()
{
  RES="${1}" ; shift
  dd bs=${BLOCKSIZE} conv=sync "$@" 2>${TMP}
  read a b < ${TMP}
  : a=$a b=$b
  if [ -n "${RES}" ] ; then
    eval ${RES}=$[${a}]
  fi
}

rmtmp ()
{
  /bin/rm -f -- ${TMP}* "$@"
}

abort ()
{
  if [ -f "$DEST" ] ; then rmtmp "$DEST" ; else rmtmp ; fi
  a=$1
  shift
  echo "$@"
  exit $a
}


##############################################################################
# Adjust the parameters in the bootable image $DEST from the arguments
make_params ()
{

  ### Tell the LOADER what LOADMODE to use ###
  if [ $LOADMODE = L ] ; then
    echo 'LOADMODE "L" is autodetected.' \
	 'Defaulting to mode "C" on a 1.4M floppy'
    LOADMODE=C
  fi

  for i in BOOTRECSIZE LOADERSIZE SYSSIZE # LOADMODE
     do echo -n $i= ; eval echo '$'$i ; done

  export BOOTRECSIZE LOADERSIZE SYSSIZE LOADMODE
  ${INSTTOOL} adjust_i386_image $DEST

}


main () {
  trap abort EXIT TERM HUP
  case "${LOADMODE}" in
    C|L)
		mkdir $(dirname $DEST) 2>/dev/null
		{
		 DD BOOTRECSIZE	< $BOOTREC
	         DD LOADERSIZE	< $LOADER
		 # adjust size for LILO loading: at least 2KB for loader !!!
		 if [ $LOADERSIZE -lt 4 ] ; then
		   dd if=/dev/zero bs=512 count=$[4-$LOADERSIZE] 2>/dev/null
		   LOADERSIZE=4
		 fi
		 cat $API $MODULES > ${TMP}xxx
                 DD SYSSIZE < ${TMP}xxx
		} > ${DEST}
		make_params
		;;
    D)		echo "Load mode Not supported yet"
		;;
    H)		echo "Load mode Not supported yet"
		;;
  esac

  altdest

  DEST=	# install successful; don't erase image
}


altdest () {
  case "x$ALTDEST" in
    x) ;;
    x/dev/*fd*)
	( cat "$DEST"
	  dd bs=3k count=1 if=/dev/zero 2>/dev/null
	) > "$ALTDEST"
	#mcopy $DEST a:
	;;
    x*) cat "$DEST" > "$ALTDEST" ;;
  esac
}

#############################################################################
config	# read the configuration
MODULES=${MODULES:-/dev/null}

ALTDEST="$1"
shift

if [ -z "$*" ]
then
  main
else
  "$*"
fi
