dnl ---------------------=|=--- pmpart.m4 ---=|=-----------------------------
dnl --=-- Protected-Mode code & data for the TUNES BIOS-based i386 API --=--
dnl
dnl --- See ~moose/info/BIOS-API.nfo for more info about this. ---
dnl
dnl
Provide({pmpart},{
dnl
dnl

pmpart

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Orig_INT_NUM	=	0x10	; first interrupt called



__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


XtdMem:		.long	0
ConvMem:	.long	0
AvConvMem:	.long	0


; Saved address for P-mode stack
PM_STACK:
PM_STACK_ofs:	.long	pmaddr(__Stack_Top__)
PM_STACK_seg:	.word	PMPARTDS

CPUType:	.word	0

RM_Part_seg:	.long	0 ; ?	; segment for RM part (long: ease of use)

rm_idt_address:
rm_idt_lim:	.word	256*4-1		; Limit of real-mode IDT
rm_idt_base:	.long	0		; Base Address of real-mode IDT

pm_idt_address:
pm_idt_lim:	.word	IDT_SIZE*8	; Limit of protected-mode IDT
pm_idt_base:	.long	pmaddr(__IDT__)	; Base Address of protected-mode IDT

})


__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


RM_Part_Phys	= allot(u,4)	; Physical address of RM Part
RM_Part_Addr	= allot(u,4)	; address of RM Part in PM Part segment
PhysicalMem	= allot(u,4)	; address of Physical Memory in PM Part segment

ReflINT_V86INT	= alallot(u,4,sizeof_V86INT)

__GDT__			= alallot(u,16,8*GDT_SIZE)
__IDT__			= alallot(u,16,8*IDT_SIZE)
__Stack__		= alallot(u,16,4*PM_STACK_SIZE)
__Stack_Top__		= allot(u,0)
__Safe_Stack__		= alallot(u,16,4*PM_STACK_SIZE)
__Safe_Stack_Top__	= allot(u,0)
K_TSS			= alallot(u,16,sizeof_TSS32)
V86_TSS = K_TSS
;Dummy_TSS		= alallot(u,16,sizeof_TSS32)
;V86_TSS		= alallot(u,16,sizeof_TSS32)


forget({
__RM_IDT__	= alallot(u,16,4*256)
})

})


__codeseg__({

NearToFar:
	STOP(fs,0xB8000,0,14,'N,0x0B)
	call [esp+8]
	STOP(fs,0xB8000,1,14,'F,0x0B)
	retf

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Descriptor tables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

LibProc({SetDescriptor},edx)
LibProc({GetDescriptor},edx)

SetGate:
; INPUT:
;   edx points to the descriptor to set
;   ebx is the offset in the descriptor for call/interrupt/trap gates
;       must be zero for task gates
;   eax contains:
;      bits 00-15=selector pointed by the gate
;      bits 16-23=counter for call gates, 0 for other gates
;      bits 24-31=P,DPL,Type
; OUTPUT:
;   descriptor set, no register lost
	mov [edx],ebx
	mov [edx+4],ebx
	mov [edx+2],eax
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

LibProc(Disable_NMI)
LibProc(Enable_NMI)

Mask_IRQs:
	outb(0xA1,#0xFF)	; mask off all slave irq's for now
	outb(0x21,#0xFB)	; mask off all master irq's but irq2 cascade
	ret

Unmask_IRQs:
	outb(0x21,#0x00)	; unmask all master irq's
	outb(0xA1,#0x00)	; unmask all slave irq's
	ret

Setup_PM_8259:
	mov cl,#IRQ_INT_BASE
	mov ch,#IRQ_INT_BASE+8
	jmp Setup_8259

Setup_RM_8259:
	mov cl,#0x08
	mov ch,#0x70
	;jmp Setup_8259

Setup_8259:
    ;;; Programming 8259A-1 (Master)
	outb(0x20,#0x11)		; initialization sequence
	outb(0x21,cl)			; start of master hardware int's
	outb(0x21,#0x4)			; 8259-1 is master
	outb(0x21,#0x1)			; 8259-1 in 8086 mode
    ;;; Programming 8259A-2 (Slave)
	outb(0xA0,#0x11)		; initialization sequence
	outb(0xA1,ch)			; start of slave hardware int's
	outb(0xA1,#0x2)			; 8259-2 is slave
	outb(0xA1,#0x1)			; 8259-2 in 8086 mode
    ;;; Done
	ret

Flush_Interrupts:
    ;;; Flush possible pending (possibly keyboard) interrupt
    ;;; Many thanks to Adam Seychell about that !
	STOP(es,0,60,9,'E,0x61)
        mov al,#0x20	; send EOI
	STOP(es,0,61,9,'O,0x61)
        out #0x20,al
	STOP(es,0,62,9,'I,0x61)
        out #0xA0,al
	STOP(es,0,63,9,'K,0x61)
	in al,#0x60	; read keyboard
	STOP(es,0,64,9,'B,0x61)
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;; Initialization routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;


Init_RM_IDT:
	mov edi,[pmaddr(PhysicalMem)]
	add edi,#IRQ_INT_BASE*4
	mov eax,[pmaddr(RM_Part_seg-2)]
	mov ax,#rmaddr(RM_IRQ)
	mov ecx,#16
	cld
  Init_RM_IDT_0:
	stosd
	add eax,#RM_IRQ_len
	loop Init_RM_IDT_0
	ret


;; enable A20, reset copro, reprogram interrupts...
;; see Linux: boot/setup.S




;; Routines to get/set an interrupt vector, that changes only the target
;; offset.

;;;; Get interrupt vector
; In:
;   BL - interrupt number
; Out:
;   EAX - 32 bit offset in code
Get_INT_Vect:
        push ebx
        movzx ebx,bl
        mov ax,[ebx*8+pmaddr(__IDT__)+6]
        shl eax,#16
        mov ax,[ebx*8+pmaddr(__IDT__)]
        pop ebx
        ret

;;;; Get interrupt vector
; In:
;   BL - interrupt number
;   EAX - 32 bit offset in code
; Out:
Set_INT_Vect:
        push ebx
        movzx ebx,bl
        mov [ebx*8+pmaddr(__IDT__)],ax
        shr eax,#16
        mov [ebx*8+pmaddr(__IDT__)+6],ax
        pop ebx
        ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;; Exception Handling ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Uncaught Exception
Uncaught_Exception:
   ;;; Show that an error occurred
	pushm(es,ds,fs,gs)
	push #PMPARTDS
	pop ds
	push ds
	pop es
	push #ALLMEMDS
	pop fs
	cld
	mov edi,[pmaddr(Video_mem_base)]
	mov [edi],#0x8F458F55  ; blinking UE
	add edi,#4
	mov ah,#0x0F
	mov ecx,#2
	call DumpHex
   ;;; Dump interrupted registers
	mov dword [pmaddr(DumpLine)],#1
	call DumpINTSTACK
	___die___ ; Uncaught Exception !

Division_Error_Trap:
	pushad
	mov dl,#0x0
	jmp Uncaught_Exception

Single_Step_Trap:
	pushad
	mov dl,#0x1
	jmp Uncaught_Exception

NMI_Trap:
	pushad
	mov dl,#0x3
	jmp Uncaught_Exception

BreakPoint_Trap:
	pushad
	mov dl,#0x3
	jmp Uncaught_Exception

Overflow_Trap:
	pushad
	mov dl,#0x4
	jmp Uncaught_Exception

Out_of_Bound_Trap:
	pushad
	mov dl,#0x5
	jmp Uncaught_Exception

Invalid_Opcode_Trap:
	pushad
	mov dl,#0x6
	jmp Uncaught_Exception

Emulate_Coproc_Trap:
	pushad
	mov dl,#0x7
	jmp Uncaught_Exception

Double_Fault_Stop:
	pushad
	mov dl,#0x8
	jmp Uncaught_Exception

Coproc_Segment_Stop:
	pushad
	mov dl,#0x9
	jmp Uncaught_Exception

Invalid_TSS_Fault:
	pushad
	mov dl,#0xA
	jmp Uncaught_Exception

Segment_Not_Present_Fault:
	pushad
	mov dl,#0xB
	jmp Uncaught_Exception

Stack_Fault:
	pushad
	mov dl,#0xC
	jmp Uncaught_Exception


;;; General protection violation Handler
General_Protection_Fault:
	zzzz({
	   push fs
	   push #ALLMEMDS
	   pop fs
	   STOP(fs,0xB8000,0,17,'G,0x0F)
	   STOP(fs,0xB8000,1,17,'P,0x0F)
	   pop fs
	})
	test byte [esp+INTSTACK_EFLAGS+2],#2 ; VM bit set ?
	bne V86_GP_Handler
        pushad
	mov dl,#0xD
	br Uncaught_Exception


Page_Fault:
	pushad
	mov dl,#0xE
	br Uncaught_Exception

Reserved_0F_Fault:
	pushad
	mov dl,#0xF
	br Uncaught_Exception

Coprocessor_Error_Fault:
	pushad
	mov dl,#0x10
	br Uncaught_Exception

Alignment_Check_Fault:
	pushad
	mov dl,#0x11
	br Uncaught_Exception




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; IRQ Handling ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Uncaught IRQ
Uncaught_IRQ:
   ;;; Show that an error occurred
	pushm(es,ds,fs,gs)
Uncaught_IRQ__:
	push #PMPARTDS
	pop ds
	push ds
	pop es
	push #ALLMEMDS
	pop fs
	cld
	mov edi,[pmaddr(Video_mem_base)]
	mov [edi],#0x8F538F49  ; blinking IR
	add edi,#4
	mov ah,#0x0F
	mov ecx,#2
	call DumpHex
   ;;; Dump interrupted registers
	mov dword [pmaddr(DumpLine)],#1
	call DumpINTSTACK
	___die___ ; Uncaught IRQ !



define({DefReflIRQ},
{
ReflIRQ_$2:
	pushm(es,ds,fs,gs,ad)
	pushm(ss,ss,#ALLMEMDS)
	popm(ds,es,fs)
	STOP(fs,0xB8000,(64+0x$2),0,'$2,0xE1)
	mov cl,#$1+0x$2
	mov ebx,#pmaddr(ReflINT_V86INT)
	call Call_INT_R86
	STOP(fs,0xB8000,(64+0x$2),0,'$2,0x61)
	popm(es,ds,fs,gs,ad)
	iretd
})



IRQ0_Timer:
	DefReflIRQ(0x08-0,0)

IRQ1_Keyboard:
	DefReflIRQ(0x08-0,1)

IRQ2_Cascade:
	DefReflIRQ(0x08-0,2)

IRQ3:
	DefReflIRQ(0x08-0,3)

IRQ4:
	DefReflIRQ(0x08-0,4)

IRQ5:
	DefReflIRQ(0x08-0,5)

IRQ6:
	DefReflIRQ(0x08-0,6)

IRQ7:
	DefReflIRQ(0x08-0,7)

IRQ8_RTC:
	DefReflIRQ(0x70-8,8)

IRQ9:
	DefReflIRQ(0x70-8,9)

IRQ10:
	DefReflIRQ(0x70-8,A)

IRQ11:
	DefReflIRQ(0x70-8,B)

IRQ12:
	DefReflIRQ(0x70-8,C)

IRQ13:
	DefReflIRQ(0x70-8,D)

IRQ14:
	DefReflIRQ(0x70-8,E)

IRQ15:
	DefReflIRQ(0x70-8,F)


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Simple V86 Handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; Ideas (and small bits of code) stolen from (in decreasing order of interest):
; - OOF/VMI by Andras Zsoter,
; - START32 and PMODE by Tran,
; - 386POWER by Lorenzo Micheletto,
; - Linux by Linus Torvalds & Co.


;;; Interface from PM/LLL to V86 mode

; (!) we assume ds,es,ss point to PMPARTDS
; debugging assumes fs points to ALLMEMDS

DefWord({Call_INT_V86})
	Enter
	PopData(ecx)
	call Call_INT_V86
	Next

DefWord({Call_V86})
	Enter
	call Call_V86
	Next

Call_INT_V86:
; cl: interrupt number
; ebx: V86INT structure
	pushad
	mov dx,#rmaddr(V86_EXIT)
	call Prepare_V86INT_for_INT
   ;; Setup V86 EFLAGS: VM, NT, IOPL=3, CLI, CLD
	mov dword [ebx+V86INT_EFLAGS],#0x00027000
	popad

Call_V86:
; ebx: V86INT structure
   ;; Save PM/LLL state
	pushm(ad,fs,gs,fd,
		[pmaddr(V86_TSS+TSS32_SS_CPL0)],
		[pmaddr(V86_TSS+TSS32_ESP_CPL0)])
   ;; Critical section
	cli
	STOP(fs,0xB8000,0,18,'C,0x0F)
   ;; Setup stack for V86 supervisor as current stack
	mov [pmaddr(V86_TSS+TSS32_ESP_CPL0)],esp
	mov [pmaddr(V86_TSS+TSS32_SS_CPL0)],ss
	STOP(fs,0xB8000,1,18,'a,0x0F)
   ;; Build a stack frame from the parameter V86INT structure
	sub esp,#sizeof_V86INT
	mov ecx,#sizeof_V86INT/4
	mov esi,ebx
	mov edi,esp
	rep
	movsd
	or byte [esp+V86INT_EFLAGS+2],#2 ; set VM bit
	STOP(fs,0xB8000,2,18,'l,0x0F)
	mov ebx,esp
	call DumpV86INT
	STOP(fs,0xB8000,3,18,'l,0x0F)
   ;; Jump into V86 !!!
	popad
	STOP(fs,0xB8000,4,18,'_,0x0F)
	STOP(fs,0xB8000,5,18,'V,0x0F)
	STOP(fs,0xB8000,6,18,'8,0x0F)
	STOP(fs,0xB8000,7,18,'6,0x0F)
	push #0x3000
	popfd
	z__z
	___die___
	iretd

   ;; Back we are !!!
Return_From_V86:
	zzzz({
	   push #ALLMEMDS
	   pop fs
	})
	STOP(fs,0xB8000,0,19,'R,0x0F)
	STOP(fs,0xB8000,1,19,'e,0x0F)
	STOP(fs,0xB8000,2,19,'t,0x0F)
	STOP(fs,0xB8000,3,19,'u,0x0F)
   ;; Initialize ds and es to PMPART
	push #PMPARTDS
	pop ds
	push ds
	pop es
	STOP(fs,0xB8000,5,18,'V,0x0F)
   ;; Pop the V86INT stack frame into the V86INT structure
	mov edi,[esp+sizeof_V86INT+PUSHAD_EBX]
	mov esi,esp
	mov ecx,sizeof_V86INT/4
	cld
	rep
	movsd
	mov esp,esi
	STOP(fs,0xB8000,6,18,'8,0x0F)
   ;; Restore PM/LLL state
	popm(ad,fs,gs,fd,
		[pmaddr(V86_TSS+TSS32_SS_CPL0)],
		[pmaddr(V86_TSS+TSS32_ESP_CPL0)])
	STOP(fs,0xB8000,7,18,'6,0x0F)
   ;; Now we can continue PM/LLL computations
	ret




;;; Supervisor Services for V86 box
V86_SERVER:
	jmp Return_From_V86



;;; General protection violation handler for V86 mode
V86_GP_Handler:
	add esp,#4	; remove error code
	pushad		; push registers; the frame is now a V86INT structure
	push #ALLMEMDS	; edit real memory
	pop fs
	STOP(fs,0xB8000,2,17,'V,0x8F)
	cli
	pushm(ss,ss)
	popm(ds,es)
	mov ebx,esp
	call DumpV86INT
	STOP(fs,0xB8000,3,17,'8,0x8F)
	___die___
   ;; Have ebx point to the CS_IP linear address
   ;; while ecx contains the IP
	movzx ebx,word [esp+V86INT_CS]
	shl ebx,#4
	movzx ecx,word [esp+V86INT_EIP]
	add ebx,ecx
	inc ecx
	mov [esp+V86INT_EIP],cx
	mov al,[ebx]
   ;; Is it an INT3 ?
	mov dl,#3
	cmp al,#0xCC
	je V86_INT_n
   ;; Is it an INTO ?
	mov dl,#4
	cmp al,#0xCE
	je V86_INT_n
   ;; Is it an INT xx ?
	cmp al,#0xCD
	je V86_INT
   ;; Last case: uncaught exception :(  ):
	mov dl,#0xC
	br Uncaught_Exception


   ;; INT instruction: get the number
   V86_INT:
	inc word [esp+32]
	mov dl,[ebx+1]
	cmp dl,#V86_SERVER_INT
	jnz V86_INT_n

   ;; Is it the SERVER being called ?
	cmp ebx,#rmaddr(V86_REQUEST)
      V86_REQUEST_Adjust:	; this address-4 should be corrected...
	jz V86_SERVER

   ;; Is it the EXIT being called ?
	cmp ebx,#rmaddr(V86_EXIT)
      V86_EXIT_Adjust:		; this address-4 should be corrected...
	beq Return_From_V86

   ;; Do interrupt for V86 task
   V86_INT_n:
	clr(ebx)
	mov bl,dl
	shl ebx,2
	movzx edx,word [esp+V86INT_SS]
	shl edx,4
	movzx ecx,word [esp+V86INT_ESP]
	sub cx,#6			; beware: overflow possible !!!
	mov word [esp+V86INT_ESP],cx
	add edx,ecx
	mov ax,[ebx+V86INT_EFLAGS]
	mov [edx+4],ax
	mov ax,[esp+V86INT_CS]
	mov [edx+2],ax
	mov ax,[esp+V86INT_EIP]
	mov [edx],ax
	and byte [esp+V86INT_EFLAGS+1],#0xFC ; disable interrupts
	mov eax,[ebx]	; interrupt vector
	mov [esp+V86INT_EIP],ax
	shr eax,16
	mov [esp+V86INT_CS],ax
	popad
	iretd


Prepare_V86INT_for_INT:
; Input:
; ebx: offset of V86INT structure
; ecx: interrupt number
; edx: offset of exit routine
; fs: RM IDT
; can scramble all registers (but esp)
   ;; Get V86 stack
	mov edi,[ebx+V86INT_SS]
	shl edi,#4
	add edi,[pmaddr(PhysicalMem)]
   ;; Reserve space on it
	movzx eax,word [ebx+V86INT_ESP]
	sub ax,#10
	add edi,eax
	mov [ebx+V86INT_ESP],ax
   ;; "Push" old CS:IP address on V86 stack
	mov eax,[ebx+V86INT_CS-2]
	mov ax,[ebx+V86INT_EIP]
	mov [edi+6],eax
   ;; "Push" interrupt return address to [RV]86_EXIT on V86 stack
	mov eax,[pmaddr(RM_Part_seg)-2]
	mov ax,dx
	mov [edi],eax
	mov word [edi+4],#0x3000 ; IOPL=3, CLI, CLD
   ;; Set Target CS:IP to interrupt routine from normal real-mode IDT
	movzx eax,cl
	seg fs
	mov eax,[eax*4]
	mov [ebx+V86INT_EIP],ax
	shr eax,16
	mov [ebx+V86INT_CS],ax
   ;; There we are !
	ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;; Simple R86 Handler ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; mode R86 is to call back between PM and Real Real Mode, *not* V86
; another difference is that R86 is invoked with fs=ALLMEMDS and gs=PMPARTDS
; though nobody prevents the RM code to trash fs and gs
; current version inspired by V86 code

;;; Interface from PM/LLL to R86 mode

DefWord({Call_INT_R86})
	Enter
	PopData(ecx)
	call Call_INT_R86
	Next

DefWord({Call_R86})
	Enter
	call Call_R86
	Next


Call_INT_R86:
; cl: interrupt number
; ebx: V86INT structure
	pushad
	mov dx,#rmaddr(R86_EXIT)
	call Prepare_V86INT_for_INT
   ;; Setup R86 EFLAGS: CLI, CLD
	mov [ebx+V86INT_EFLAGS],#0x7000
	popad

Call_R86:
	STOP(fs,0xB8000,0,24,'R,0x07)
   ;; save PM/LLL context
	pushm(ad,fs,gs,fd)
   ;; begin critical section
	cli
	;call Disable_NMI
	;STOP(fs,0xB8000,1,24,'C,0x07)
   ;; Save PM stack address
	mov [pmaddr(PM_STACK_ofs)],esp
   ;; "Push" EIP,CS,FD on RM stack
	clr(ecx)
	mov cx,[ebx+V86INT_ESP]
	sub cx,#12
	clr(edi)
	mov di,[ebx+V86INT_SS]
	shl edi,#4
	add edi,ecx
	add edi,[pmaddr(PhysicalMem)]
	;STOP(fs,0xB8000,2,24,'S,0x07)
	mov eax,[ebx+V86INT_EIP]
	mov [edi],eax
	mov eax,[ebx+V86INT_CS]
	mov [edi+4],eax
	mov eax,[ebx+V86INT_EFLAGS]
	mov [edi+8],eax
	;STOP(fs,0xB8000,3,24,'S,0x07)
   ;; Extract Restore_State parameters from V86INT structure
	push [ebx+V86INT_SS]
	push ecx
	sub esp,#sizeof_PUSHAD+4*4
	mov edi,esp
	mov esi,ebx
	mov ecx,#8
	rep
	movsd
	add esi,#V86INT_ES-V86INT_EIP
	rept(4,{
	   movsd})
	;STOP(fs,0xB8000,4,24,'V,0x07)
   ;; initialize fs,gs to ALLMEM, PMPART
	pushm(#ALLMEMDS,#PMPARTDS)
	popm(fs,gs)
	;STOP(fs,0xB8000,5,24,'G,0x07)
   ;; Jump into R86 mode
	lidt [pmaddr(rm_idt_address)]
	;STOP(fs,0xB8000,6,24,'I,0x07)

   ;; debug.....
	;movzx eax,word [pmaddr(rm_idt_lim)]
	;mov ebx,dword [pmaddr(rm_idt_base)]
	;z__z

	jmpi #rmaddr(GoReal), #RMPARTCS

Return_From_R86:
   ;; Initialize ds,es to PMPART
	push #PMPARTDS
	pop ds
	push ds
	pop es
	zzzz({
	  push #ALLMEMDS
	  pop fs
	})
	STOP(fs,0xB8000,0,23,'R,0x0B)
   ;; Compute logical address of RM stack
	clr(ebx)	; bx is also cleared for below
	clr(esi)
	mov bx,sp
	mov si,ss
	shl esi,#4
	add esi,ebx
	add esi,[pmaddr(PhysicalMem)]

	STOP(fs,0xB8000,1,23,'e,0x0B)
	STOP(fs,0xB8000,2,23,'t,0x0B)
	zzzz({mov [pmaddr(DumpLine)],#1})
	STOP(fs,0xB8000,3,23,'u,0x0B)

   ;; Restore PM stack
	lss esp,[pmaddr(PM_STACK)]
	STOP(fs,0xB8000,4,23,'r,0x0B)
	STOP(fs,0xB8000,5,23,'n,0x0B)
	STOP(fs,0xB8000,6,23,'_,0x0B)

   ;; Copy it to parameter V86INT structure
	mov edi,[esp+PUSHAD_EBX]

	mov ebx,edi
	call DumpV86INT
	STOP(fs,0xB8000,7,23,'F,0x0B)

	mov ecx,#sizeof_V86INT/4
	cld
	rep
	movsd
	STOP(fs,0xB8000,8,23,'r,0x0B)

	call DumpV86INT
	STOP(fs,0xB8000,9,23,'o,0x0B)

	z__z
	STOP(fs,0xB8000,10,23,'m,0x0B)

   ;; End Critical Section
	;call Enable_NMI
	lidt [pmaddr(pm_idt_address)]
	STOP(fs,0xB8000,11,23,'_,0x0B)
   ;; Restore PM/LLL state
	popm(ad,fs,gs,fd)
	STOP(fs,0xB8000,12,23,'R,0x0B)
	STOP(fs,0xB8000,13,23,'8,0x0B)
	STOP(fs,0xB8000,14,23,'6,0x0B)
   ;; Now we can continue PM/LLL computations
	ret




;;;;;;;;;;;;;;;;;;;;;;;;;;;; Simple I/O subroutine ;;;;;;;;;;;;;;;;;;;;;;;;;;;;


DumpHex:
;;Input:
;; edx: hex number to dump
;; ecx: number of hex digits
;; edi: video buffer position
;; ah: video attribute
;;Output:
;; every register saved
;; edi: ends right after the last character
	lea edi,[edi+ecx*2]
pushing({eax,ecx,edx,edi},{
   Loopno0({
	dec edi
	dec edi
	mov al,dl
	and al,#0x0F
	add al,#0x30
	Ifcompare(ae,al,#0x3A,{add al,#0x41-0x3A})
	mov [edi],ax
	shr edx,#4
})})
	ret




forget({

;;;;;;;;;;;;;;; Standard objects (their low_level implementation) ;;;;;;;;;;;;

;;; Stdin as a byte per byte input ;;;
__stdin_param__:	.word	0 ; ?
__stdin_Getc__:		.word	cGetc
__stdin_Flush__:	.word	False
__stdin_EOF__:		.word	False
;;; Stdout as a byte per byte output ;;;
__stdout_param__:	.word	0 ; ?
__stdout_Putc__:	.word	cPutc
__stdout_Flush__:	.word	False
__stdout_FULL__:	.word	False

Putc:
	push si
	mov si,[rmaddr(__stdout_param__)]
	call [rmaddr(__stdout_Putc__)]
	mov [rmaddr(__stdout_param__)],si
	pop si
	ret

Getc:
	push si
	mov si,[rmaddr(__stdout_param__)]
	call [rmaddr(__stdout_Putc__)]
	mov [rmaddr(__stdout_param__)],si
	pop si
	ret

})



})

})dnl --- end of pmpart.m4
dnl
