dnl ---------------------=|=--- rmpart.m4 ---=|=-----------------------------
dnl --=-- Real-Mode code & data for the TUNES BIOS-based i386 API --=--
dnl
dnl --- See ~moose/info/BIOS-API.nfo for more info about this. ---
dnl
dnl
Provide({rmpart},{
dnl
dnl



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


rmpart

__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


PM_Target:		; Address to jump to when entering PM
	PM_Target_ofs:	.long	pmaddr(PM_Entry_Point)
	PM_Target_seg:	.word	PMPARTCS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PM_GDT_address:		; Length & Base address of protected-mode GDT
	PM_GDT_lim:	.word	Orig_GDT_Len-1
	PM_GDT_base:	.long	Orig_GDT

})



__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

alignheap(ru,4)
; (!) StackTop mustn't be at the end, or it will be 64K, and
; esp=0x10000 is very bug-prone when going RM to PM !!!!
RMPartStackBot	= allot(ru,256*2)
RMPartStackTop	= allot(ru,0)

RM_cr0		= allot(ru,4)	; value of cr0 for entering real mode
PM_cr0		= allot(ru,4)	; value of cr0 for entering protected mode

;RM_STACK 	= alallot(ru,4,0) 	; Saved addr for R-mode stack
;RM_STACK_ofs	= allot(ru,2)
;RM_STACK_seg:	= allot(ru,2)

})


__codeseg__({
pushdef({DumpERegs},{RM_DumpERegs})dnl

;;;;;;;;;;;;;;;;;;;;;;;;;; Standard Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Puts_:
	call Putc
Puts:
; Adapted from MicroSoft's own boot sector
; DS:SI=ASCIIZ string to print
; AX spoilt, SI points just after the string-ending zero, 7 put into BX.
	cld
	lodsb
	or al,al
	jnz Puts_
	ret

Beep:
; Echo a BEEP to the the Console
	mov al,#7
	;jmps cPutc

cPutc:
; Outputs a character to console
; spoils AX and DI, keeps si
	pushing({ad,f,es},{
	mov ah,#0xE	; function Teletype
	mov bx,#7	; attribute
	int #0x10	; Video BIOS call
	})
Return_:
	ret

cGetc:
	pushing(f,{
	clr(ah)
	int #0x16
	})
	ret

forget({
; the true condition is "c"; false is "nc"
True:
	stc
	ret

False:
	clc
	ret
})

Putc=cPutc
Getc=cGetc










;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pushdef({__IO_Delay__},{call Delay})dnl

;;; Delay between I/O, as in Linux.
Delay:
	jmp DelayRet
DelayRet:
	ret

RM_Mask_IRQs:
	outb(0xA1,#0xFF)	; mask off all interrupts for now
	outb(0x21,#0xFB)	; mask all irq's but irq2 which is cascaded
	ret

RM_Unmask_IRQs:
	outb(0x21,#0x00)
	outb(0xA1,#0x00)
	ret

RM_Disable_NMI:
	LibInline(Disable_NMI)
	ret

RM_Enable_NMI:
	LibInline(Enable_NMI)
	ret

Reset_Copro:
	clr(ax)
	out 0xF0,al
	call Delay
	out 0xF1,al
	ret

LibProc(KB_Ready,Timeout_8042)

; What to do in case of time out from 8042
Timeout_8042:
	jmp KB_Ready

Disable_A20:
	mov ah,#set_A20_OFF
	jmp Set_A20

Enable_A20:
	mov ah,#set_A20_ON
	;jmp Set_A20

LibProc(Set_A20)


forget({
RM_Setup_PM_8259:
	mov cx,#IRQ_INT_BASE+256*(IRQ_INT_BASE+8)
	jmp RM_Setup_8259

RM_Setup_RM_8259:
	mov cx,#0x7008
	;jmp RM_Setup_8259

RM_Setup_8259:
    ;;; Programming 8259A-1 (Master)
	outb(0x20,#0x11)		; initialization sequence
	outb(0x21,cl)			; start of master hardware int's
	outb(0x21,#0x4)			; 8259-1 is master
	outb(0x21,#0x1)			; 8259-1 in 8086 mode
    ;;; Programming 8259A-2 (Slave)
	outb(0xA0,#0x11)		; initialization sequence
	outb(0xA1,ch)			; start of slave hardware int's
	outb(0xA1,#0x2)			; 8259-2 is slave
	outb(0xA1,#0x1)			; 8259-2 in 8086 mode
    ;;; Done
	ret

RM_Flush_Interrupts:
    ;;; Flush possible pending (possibly keyboard) interrupt
    ;;; Many thanks to Adam Seychell about that !
	STOP(es,0,60,9,'E,0x61)
        mov al,#0x20	; send EOI
	STOP(es,0,61,9,'O,0x61)
        out #0x20,al
	STOP(es,0,62,9,'I,0x61)
        out #0xA0,al
	STOP(es,0,63,9,'K,0x61)
	in al,#0x60	; read keyboard
	STOP(es,0,64,9,'B,0x61)
	ret
})

popdef({__IO_Delay__})


;;;;;;;;;;;;;;;;;;;;;;;;;;; Support for V86 Mode ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

V86_EXIT:
; - must be far called, or push cs; near called, for state to be properly
;   saved
	int #V86_SERVER_INT
	retf


V86_REQUEST:
; - must be far called, or push cs; near called, for state to be properly
;   saved
	int #V86_SERVER_INT
	retf


;;;;;;;;;;;;;;;; Calling 32-bit routines from 16-bit PM ;;;;;;;;;;;;;;;;;;;;;

define({CallPMsub},
       {push dword #pmaddr($1)
	call RM_NearToFar})dnl

RM_NearToFar:
	push dword #PMPARTCS
	push dword #pmaddr(NearToFar)
	push dword [esp+10]
	calli dword [esp+4]
	lea esp,[esp+12]
	ret 4


;;;;;;;;;;;;;;;;;;;;;;;;;;; Support for R86 Mode ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;; Save State for R86 ;;;;;
R86_EXIT:
Save_State:
; - must be far called, or push cs; near called, for state to be properly
;   saved
; - Saves all register for use by Return_From_R86
; - jumps into protected mode
	pushm(fd,esp,dword ss,dword es,dword ds,dword fs,dword gs)
	movzx esp,sp
	push word #0
	push word [esp+2+7*4+2]
	push word #0
	push word [esp+2+8*4]
	add [esp+12],#6*4
	pushad
	;jmp GoProtect

;;;;; Real Mode to Protected Mode switcher ;;;;;
GoProtect:
; PREcondition
; * cs points to the rmpart
; POSTcondition
; * A20 line is enabled
; * all hardware interrupts are disabled
; * Processor enters Protected mode
; * Processor jumps into 32-bit far address contained in
; * eax,ebx,ecx,edx may be scrambled
; * ds moved to RM_Part_Seg
; * es,ss,fs,gs unchanged (internal values also)
; * esp,ebp,esi,edi unchanged

	push cs
	pop ds
   zwoz({{"Going PM, Target="}},[rmaddr(PM_Target_seg)],{":"})
   zdwoz({},[rmaddr(PM_Target_ofs)],{CRLF})
	cli
   ;__("Disabling NMI",CRLF)
	  ;call RM_Disable_NMI
   __("Enabling A20",CRLF)
	call Enable_A20
   zdwoz({"Loading GDT at "},[rmaddr(PM_GDT_base)],{CRLF})
	zzzz({call WaitGoPM})
	lgdt [rmaddr(PM_GDT_address)]
	mov eax,[rmaddr(PM_cr0)]	; 32 bit msw
	test byte [rmaddr(GP_CR0)],#1
	mov cr0,eax			; This is it!
 GP_CR0:
	jmpi dword [rmaddr(PM_Target)]



;;;;; Protected to Real Mode switcher ;;;;;

GoReal:
; PREconditions:
; * Interrupts are disabled,
; POSTconditions:
; * (just) do the switching to Real mode and jump to 16-bit far [RM_Target]
; NOTE:
; I had problems having this code in a 32 bit segment,
; as the final jump seemed to get nowhere (or get it in real 32-bit mode ?)
	zzzz({
	  mov ax,#ALLMEMDS
	  mov fs,ax
	})
	STOP(fs,0xB8000,3,22,'G,0x0F)
	STOP(fs,0xB8000,1,15,'G,0x0C)
	seg cs
	mov eax,[rmaddr(RM_cr0)]	; 16 bit real-mode msw
	STOP(fs,0xB8000,2,15,'o,0xC)
	zzzz({call WaitGoRM})
	seg cs
	test byte [rmaddr(GR_CR0)],#1 ; put following instr. in cache...
	mov cr0,eax		; Back to the Real World :( ):
   GR_CR0:
	flushinstr
	STOP(fs,0xB8000,3,15,'R,0xC)

	.byte 0xEA	;seg cs	;jmpi word [RM_Target]
RM_Target:
	RM_Target_ofs:	.word	rmaddr(Restore_State)
	RM_Target_seg:	.word	0 ; ??? Will be [RM_Part_seg]


;;;;; Restore State from R86 ;;;;;
; * Parameter is:
;  - the zero-padded values for es,ds,fs,gs
;  - a pushad/popad frame
;  - the value to iretd
; * Restore_State_fg does restore fs and gs;
; * Restore_State does not restore fs and gs, and instead, expects
;  fs=ALLMEM and gs=PMPART

Restore_State_fg:
	popad
	popm(dword es,dword ds,dword fs,dword gs)
	lss esp,[esp]
	iretd

Restore_State:
	zzzz({
	  seg cs
	  mov word [rmaddr(RM_DumpLine)],#18
	})
	STOP(fs,0xB8000,0,0,'R,0x61)
	popad
	STOP(fs,0xB8000,1,0,'S,0x61)
	popm(dword es,ds)
	STOP(fs,0xB8000,2,0,'8,0x61)
	;zWZonez(,[esp+4+10+sizeof_PUSHA+2],[esp+10+sizeof_PUSHA+2],16,{CRLF})
	STOP(fs,0xB8000,3,0,'6,0x61)
	lss esp,[esp+10]
	STOP(fs,0xB8000,4,0,'D,0x61)
	;rz__z
	STOP(fs,0xB8000,5,0,'Z,0x61)
	;___die___
	iretd


;;;;;;;;;;;;;;;;;;; Hook from hardware IRQ to BIOS handlers ;;;;;;;;;;;;;;;;;;
;;; (when MOOSE I/O control is disabled)

RM_IRQ:

define({DefEmulIRQ},
{RM_IRQ_$2:
	STOP(fs,0xB8000,(48+0x$2),0,'$2,0x8F)
	int #$1+0x$2
	STOP(fs,0xB8000,(48+0x$2),0,'$2,0x0F)
	iret
})

irp(X,{0,1,2,3,4,5,6,7},{DefEmulIRQ(0x08-0,X)})
irp(X,{8,9,A,B,C,D,E,F},{DefEmulIRQ(0x70-8,X)})
RM_IRQ_len = RM_IRQ_1 - RM_IRQ_0


})

forget({
; one way to do things is to intercept everything through our one
; RM IDT outside of conventional memory which points to this routine,
; with low byte of CS being the interrupt number,
; except perhaps for hardware interrupts.

INT_Interceptor:
	push cs
	.byte 0xEA ; jmpi
Actual_INT_Interceptor:
	.word rmaddr(INT_Interceptor_0)
Actual_INT_Interceptor_seg:
	.word 0 ; ?
})

popdef({DumpERegs})dnl

})dnl --- end of rmpart.m4
dnl
