#include <stdio.h>  /* for sscanf () */
#include <fcntl.h>  /* for open () */
#include <unistd.h> /* for truncate (), lseek () */
#include <errno.h>  /* for error number */
#include <stdlib.h> /* for exit() */
#include <string.h> /* for strcmp () */
#include <bytesex.h>

/* Simple installation tool */

/*
* We should eventually:
* - integrate it to the more generic OTOP sources
*/



#undef DEBUG
#ifdef DEBUG
int __DBG__=1 ;
#  define DBG(x)	(__DBG__?(x):0)
#else
#  define DBG(x)
#endif


#if   __BYTE_ORDER==1234
#  undef	SWAP_BYTES
#elif __BYTE_ORDER==4321
#  define	SWAP_BYTES
#else
#  error Unknown bytesex !
#endif


typedef unsigned long		dword ;
typedef unsigned short		word ;
typedef unsigned char		byte ;


#define TRY0(x,bad)	(((x)==(bad))?exit(errno):0)
#define TRY(x)		TRY0(x,-1L)

#define CHECK(x,good,msg)   (((x)==(good))?0:(fprintf(stderr,msg "\n"),exit(10)))


#define BSIZE 16384
unsigned char BUFFER [BSIZE] ;

#define SECTSIZE 512



#define putLEbyte putbyte
#define putBEbyte putbyte
#define getLEbyte getbyte
#define getBEbyte getbyte
inline void putbyte(void*x,byte y) {
  *((byte*)x)=y ;
}
inline byte getbyte(void*x) {
  return *((byte*)x) ;
}

inline void putLEword(void*x,word y) {
  ((byte*)x)[0]=(byte)y ;
  ((byte*)x)[1]=(byte)(y>>8) ;
}
inline void putLEdword(void*x,dword y) {
  ((byte*)x)[0]=(byte)y ;
  ((byte*)x)[1]=(byte)(y>>8) ;
  ((byte*)x)[2]=(byte)(y>>16) ;
  ((byte*)x)[3]=(byte)(y>>24) ;
}
inline word getLEword(void*x) {
  return  ((byte*)x)[0] | (((byte*)x)[1]<<8) ;
}
inline dword getLEdword(void*x) {
  return    ((byte*)x)[0]      | (((byte*)x)[1]<<8) 
         | (((byte*)x)[2]<<16) | (((byte*)x)[3]<<24) ;
}

inline void putBEword(void*x,word y) {
  ((byte*)x)[1]=(byte)y ;
  ((byte*)x)[0]=(byte)(y>>8) ;
}
inline void putBEdword(void*x,dword y) {
  ((byte*)x)[3]=(byte)y ;
  ((byte*)x)[2]=(byte)(y>>8) ;
  ((byte*)x)[1]=(byte)(y>>16) ;
  ((byte*)x)[0]=(byte)(y>>24) ;
}
inline word getBEword(void*x) {
  return  ((byte*)x)[1] | (((byte*)x)[0]<<8) ;
}
inline dword getBEdword(void*x) {
  return    ((byte*)x)[3]      | (((byte*)x)[2]<<8) 
         | (((byte*)x)[1]<<16) | (((byte*)x)[0]<<24) ;
}



/**** OLD STUFF

#ifdef SWAP_BYTES
#  define gooddword(x) (swapbytes(x))
#  define filedword(x) (swapbytes(x))
#  define goodword(x) (wswapbytes(x))
#  define fileword(x) (wswapbytes(x))
#else
#  define gooddword(x) (x)
#  define filedword(x) (x)
#  define goodword(x) (x)
#  define fileword(x) (x)
#endif
inline void reversebytes(char*x,int i) {
  int j ;
  char t ;
  for(j=0;j+j<i;j++) {
    t=x[j] ;
    x[j]=x[i-j] ;
    x[i-j]=t ;
  }
}
inline dword swapbytes(dword x) {
  x = (x<<16) | (x>>16) ;
  x = ((x&0x00FF00FF)<<8) | (x&0xFF00FF00>>8) ;
  return x ;
}
inline word wswapbytes(word x) {
   x = (x<<8) | (x>>8) ;
   return x ;
}

long filesize(int fd)
{
  long l ;
  TRY(l = lseek(fd,0,SEEK_END)) ;
  return l ;
}

******/



dword LEchksum(int fd,long beg,long len)
{
  long Z,i,j,L ;
  dword S=0 ;


  TRY(  lseek (fd,beg,SEEK_SET)  );

  for(i=0;i+BSIZE<len;i+=BSIZE) {
    TRY(  read (fd,BUFFER,BSIZE)  );
    for(j=0;j<BSIZE;j+=4) {
      S += getLEdword(BUFFER+j) ;
    }
  }


  L=len-i ;
  TRY(  read (fd,BUFFER,L)  );
  memset(BUFFER+L,BSIZE-L,0) ;
    for(j=0;j<L;j+=4) {
      S += getLEdword(BUFFER+j) ;
    }

  return S ;
}



#ifdef DEBUG
dword partLEchksum(int fd,long beg,long len)
{
  long Z,i,j,k,L ;
  dword S=0 ;


  TRY(  lseek (fd,beg,SEEK_SET)  );

  for(i=0;i+BSIZE<len;i+=BSIZE) {
    TRY(  read (fd,BUFFER,BSIZE)  );
    for(k=0;k<BSIZE;k+=SECTSIZE) {
      for(j=0;j<SECTSIZE;j+=4) {
	S += getLEdword(BUFFER+k+j) ;
      }
      fprintf(stderr,"Partial Checksum 0x%08lX.\n",S) ;
    }
  }

  L=len-i ;
  TRY(  read (fd,BUFFER,L)  );
  memset(BUFFER+L,BSIZE-L,0) ;

    for(k=0;k<L;k+=SECTSIZE) {
      for(j=0;j<SECTSIZE;j+=4) {
	S += getLEdword(BUFFER+k+j) ;
      }
      fprintf(stderr,"Partial checksum 0x%08lX.\n",S) ;
    }

  return S ;
}
#endif


#define writeLEbyte writebyte
#define readLEbyte readbyte
#define writeBEbyte writebyte
#define readBEbyte readbyte

void writebyte(int fd,long pos,byte b) {
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  write (fd,&b,1)  );
}
byte readbyte(int fd,long pos) {
  byte x ;
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  read (fd,&x,1)  );
  return x ;
}

void writeLEword(int fd,long pos,word w) {
  byte x[2] ;
  putLEword(x,w) ;
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  write (fd,&x,2)  );
}
void writeLEdword(int fd,long pos,dword l) {
  byte x[4] ;
  putLEdword(x,l) ;
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  write (fd,&x,4)  );
}

word readLEword(int fd,long pos) {
  byte x[2] ;
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  read (fd,&x,2)  );
  return getLEword(x) ;
}
dword readLEdword(int fd,long pos) {
  byte x[4] ;
  TRY(  lseek (fd,pos,SEEK_SET)  );
  TRY(  read (fd,&x,4)  );
  return getLEdword(x) ;
}


dword getenvint (char*A,dword def) {
  char *X=getenv(A) ;
  return X ? strtoul(X,NULL,0) : def ;
}
long getenvchar (char*A,char def) {
  char *X=getenv(A) ;
  return X ? *X : def ;
}

void adjust_i386_image (int fd) {

/*
 * SYSSIZE: size without bootrec+loader
 */

  dword BOOTRECSIZE	= getenvint("BOOTRECSIZE",0) ;
  dword LOADERSIZE	= getenvint("LOADERSIZE",0) ;
  dword SYSSIZE		= getenvint("SYSSIZE",0) ;

  char LOADMODE		= getenvchar("LOADMODE",'C') ;

  dword APIOFS		= BOOTRECSIZE+LOADERSIZE ;
  dword TOTALSIZE	= BOOTRECSIZE+LOADERSIZE+SYSSIZE ;
  dword SYSSIZE_B	= SECTSIZE*SYSSIZE ;
  dword APIOFS_B	= SECTSIZE*APIOFS ;
  dword TOTALSIZE_B	= SECTSIZE*TOTALSIZE ;

  dword checksum ;

  CHECK(BOOTRECSIZE,		1,	"Boot sector over 512 bytes !") ;
  CHECK(readLEword(fd,510),	0xAA55,	"Invalid boot sector signature.") ;
  CHECK(readLEdword(fd,APIOFS_B+56), 0,	"checksum field should be zero.") ;


  writeLEword(fd,0x00E,		TOTALSIZE) ;	/* reserved sectors */
  writebyte(fd,0x024,		LOADMODE) ;	/* Load Mode */
  writeLEword(fd,0x025,		APIOFS) ;	/* where loading shall begin */
  writebyte(fd,0x02A,		LOADERSIZE) ;	/* loader size for bootrec */
  writebyte(fd,0x1F1,		LOADERSIZE) ;	/* loader size for bootrec */
  writeLEword(fd,0x1F4,		SYSSIZE_B>>4) ; /* syssize in para for LILO */
  writeLEdword(fd,APIOFS_B+32,	SYSSIZE_B) ;	/* syssize for loader */

  checksum = LEchksum(fd,APIOFS_B,SYSSIZE_B) ;
  writeLEdword(fd,APIOFS_B+56,	-checksum) ;    /* checksum for loader */

  DBG(  printf("Checksum is 0x%08lX.\n", checksum)  ) ;
  DBG(  partLEchksum(fd,APIOFS_B,SYSSIZE_B)  ) ;


  printf("Total image size is %ld bytes (%ld sectors).\n",
	 TOTALSIZE_B,TOTALSIZE) ;
}


void report_error(void) {
  fprintf(stderr,"You shouldn't run this program manually.\n") ;
  exit(2) ;
}

typedef void (*action_t)(int fd) ;

#define namefun(x)  {#x,(action_t)x}

struct tasks {
  char *name ;
  action_t fun ;
} task [] =
{ namefun(adjust_i386_image),
  {NULL,(action_t)report_error}
} ;


int find_task(char*n) {
   int i=0 ;
   while(task[i].name) {
     if(!strcmp(task[i].name,n)) { break ;}
     else i++ ;
   }
   return i ;
}


int main(int argc, char **argv)
{
  int fd ;

  TRY0(fd = open(argv[2],O_RDWR),0) ;

  (*(task[find_task(argv[1])].fun))(fd) ;

  return 0 ;
}
